//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Useful Stuff
//   -- Header File
//  
// Primary Author: Adam Bargteil (adamb@cs.berkeley.edu)
// 
// This file contains a bunch of useful functions, and classes.
// List of doubles, integers, vectors, tirangles, etc.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// Copyright (c) 2003-2005, Regents of the University of California.  All
// rights reserved.
//
// This software is part of the Berkeley Fluid Animation & Simulation
// Toolkit.  The name "Berkeley Fluid Animation & Simulation Toolkit" is
// a trademark of the Regents of the University of California.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
//   Redistributions of source code must retain the above copyright
//  notice, this list of conditions and the following disclaimer.
//
//  Redistributions in binary form must reproduce the above copyright
//  notice, this list of conditions and the following disclaimer in the
//  documentation and/or other materials provided with the distribution.
//
//  Redistributions in binary form as an executable program, or part of
//  an executable program must, when run for the first time by a given
//  user, prominently display the above copyright notice, this list of
//  conditions and the following disclaimer.
//
//  Neither the name of the University of California, Berkeley nor the
//  names of its contributors may be used to endorse or promote products
//  derived from this software without specific prior written
//  permission.
//
//  ** Animations, still images, or other works created using this
//  ** software must clearly indicate in the list of credits that this
//  ** software was used.  The software shall be referred to as the
//  ** Berkeley Fluid Animation & Simulation Toolkit.  If the software
//  ** is included, either in whole or in part, within another software
//  ** package, credit must still be given to the Berkeley Fluid
//  ** Animation & Simulation Toolkit.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------

#ifndef bfastUtil_H
#define bfastUtil_H

#include "bfast.H"
#include <iostream>
#include <fstream>
#include "bfastVector.H"

#ifdef WIN32
#if _MSC_VER >= 1300
#include <hash_map>
#include <hash_set>
#ifndef HASH_MAP
#define HASH_MAP std::hash_map
#endif
#ifndef HASH_SET
#define HASH_SET std::hash_set
#endif
#else
#include <map>
#include <set>
#ifndef HASH_MAP
#define HASH_MAP std::map
#endif
#ifndef HASH_SET
#define HASH_SET std::set
#endif
#endif
#else
#include <ext/hash_map>
#include <ext/hash_set>
#ifndef HASH_MAP
#define HASH_MAP __gnu_cxx::hash_map
#endif
#ifndef HASH_SET
#define HASH_SET __gnu_cxx::hash_set
#endif
#endif

#ifdef INVENTOR_DEFINED
#include <Inventor/SoDB.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoOutput.h>
#include <Inventor/nodes/SoSphere.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#endif


///////////////////////////////////////////////////////////////////

class BfastTri {
public:
  int indices[3];
  inline int &operator[](const unsigned int &i) { return indices[i];};
  inline int operator[](const unsigned int &i) const { return indices[i];};
  inline void set(int x, int y, int z) {indices[0] = x; indices[1] = y; indices[2] = z;};
  inline BfastTri(int x, int y, int z) {indices[0] = x; indices[1] = y; indices[2] = z;};
  inline BfastTri() {};
  inline BfastTri &operator=(const BfastTri &that);
};
  
/////////////////////////////////////////////////////////////////

class BfastInt3 {
public:
	inline BfastInt3() { data[0] = data[1] = data[2] = 0; };
	inline BfastInt3(int i) { data[0] = data[1] = data[2] = i; };
	inline BfastInt3(int i1, int i2, int i3) { data[0] = i1; data[1] = i2; data[2] = i3; };
	inline BfastInt3(const BfastInt3 &da) { data[0] = da[0]; data[1] = da[1]; data[2] = da[2]; };
	inline BfastInt3(const int *da) { data[0] = da[0]; data[1] = da[1]; data[2] = da[2]; };

	inline int &operator[](unsigned int i) { return data[i]; };
	inline int  operator[](unsigned int i) const { return data[i]; };

	inline int &operator()(unsigned int i) { return data[i]; };
	inline int  operator()(unsigned int i) const { return data[i]; };

	inline BfastInt3 &set(int i) { data[0] = data[1] = data[2] = i; return *this; };
	inline BfastInt3 &set(int i1, int i2, int i3) { data[0] = i1; data[1] = i2; data[2] = i3; return *this; };
	
	inline BfastInt3 &set(const BfastInt3 &da) { data[0] = da[0]; data[1] = da[1]; data[2] = da[2]; return *this; };
	inline BfastInt3 &set(const int *da) { data[0] = da[0]; data[1] = da[1]; data[2] = da[2]; return *this; };
	
	inline BfastInt3 &operator=(int i) { data[0] = data[1] = data[2] = i; return *this; };
	inline BfastInt3 &operator=(const BfastInt3 &da) { data[0] = da[0]; data[1] = da[1]; data[2] = da[2]; return *this; };
	inline BfastInt3 &operator=(const int *da) { data[0] = da[0]; data[1] = da[1]; data[2] = da[2]; return *this; };

private:
	int data[3];
};

std::istream &operator>>(std::istream &strm, BfastTri &t);
std::ostream &operator<<(std::ostream &strm, const BfastTri &t);

std::istream &operator>>(std::istream &strm, std::vector<BfastTri> &p);
std::ostream &operator<<(std::ostream &strm, const std::vector<BfastTri> &p);
  
std::istream &operator>>(std::istream &strm, std::vector<BfastVector3> &p);
std::ostream &operator<<(std::ostream &strm, const std::vector<BfastVector3> &p);

std::istream &operator>>(std::istream &strm, BfastInt3 &p);
std::ostream &operator<<(std::ostream &strm, const BfastInt3 &p);

std::istream &operator>>(std::istream &strm, std::vector<BfastInt3> &p);
std::ostream &operator<<(std::ostream &strm, const std::vector<BfastInt3> &p);
  
std::istream &operator>>(std::istream &strm, std::vector<int> &p);
std::ostream &operator<<(std::ostream &strm, const std::vector<int> &p);

std::istream &operator>>(std::istream &strm, std::vector<double> &p);
std::ostream &operator<<(std::ostream &strm, const std::vector<double> &p);

namespace bfast {
	void read(std::ifstream &in, std::vector<int> &l);
	void write(std::ofstream &out, const std::vector<int> &l);
	void read(std::ifstream &in, std::vector<BfastReal> &l);
	void write(std::ofstream &out, const std::vector<BfastReal> &l);
	void read(std::ifstream &in, BfastTri &t);
	void write(std::ofstream &out, const BfastTri &t);
	void read(std::ifstream &in, std::vector<BfastTri> &l);
	void write(std::ofstream &out, const std::vector<BfastTri> &l);
	void read(std::ifstream &in, std::vector<BfastVector3> &l);
	void write(std::ofstream &out, const std::vector<BfastVector3> &l);
	void read(std::ifstream &in, std::vector<BfastInt3> &l);
	void write(std::ofstream &out, const std::vector<BfastInt3> &l);
#ifdef INVENTOR_DEFINED
	SoIndexedFaceSet *faceSet(const std::vector<BfastTri> &l);
	SoCoordinate3 *coord(const std::vector<BfastVector3> &l);
#endif

};


void randomize(std::vector<int> &list);
void computeNormals(const std::vector<BfastVector3> &meshPts, const std::vector<BfastTri> &triangles, 
										std::vector<BfastVector3> &vertexNormals, std::vector<BfastVector3> &faceNormals);
void bvtRead(char *fname, std::vector<BfastVector3> &meshPts, std::vector<BfastTri> &triangles);
void gtDump(char *fname, const std::vector<BfastVector3> &meshPts, const std::vector<BfastTri> &triangles);
void gtRead(char *fname, std::vector<BfastVector3> &meshPts, std::vector<BfastTri> &triangles);
void bgtDump(char *fname, const std::vector<BfastVector3> &meshPts, const std::vector<BfastTri> &triangles);
bool readObjFile(char *fname, std::vector<BfastVector3> &pts, std::vector<BfastTri> &triangles);


inline BfastReal absmin2 (BfastReal x, BfastReal y) {
  if (fabs(x) < fabs(y)) return x;
  return y;
}

inline BfastReal absmax2 (BfastReal x, BfastReal y) {
  if (fabs(x) > fabs(y)) return x;
  return y;
}

inline int sign(BfastReal x) {
  if (x>0.0) return 1;
	else return -1;
}

inline BfastReal max3 (BfastReal x, BfastReal y, BfastReal z) {
  if (x > y && x > z) return x;
  if (y > z) return y;
  return z;
}

inline BfastReal min3 (BfastReal x, BfastReal y, BfastReal z) {
  if (x < y && x < z) return x;
  if (y < z) return y;
  return z;
}

inline int max2 (int x, int y) {
  if (x>y) return x;
  else return y;
}

inline int min2 (int x, int y) {
  if (x<y) return x;
  else return y;
}

inline BfastReal min2 (BfastReal x, BfastReal y) {
  if (x<y) return x;
  else return y;
}

inline BfastReal max2 (BfastReal x, BfastReal y) {
  if (x>y) return x;
  else return y;
}

inline BfastReal max5(BfastReal x1, BfastReal x2, BfastReal x3, BfastReal x4, BfastReal x5) {
  BfastReal m = x1;
  if (x2 > m) m = x2;
  if (x3 > m) m = x3;
  if (x4 > m) m = x4;
  if (x5 > m) m = x5;
  return m;
}

BfastReal pow2(int p);

// trace back in time (semi-Lagrangian)
inline BfastVector3 bfastTraceBack(const BfastVector3 &x, BfastVector3 (*evalVel)(const BfastVector3 &x, void *data),
																	 BfastReal dt, void *evalVelData=NULL) {
	BfastVector3 v = evalVel(x,evalVelData);
	BfastVector3 y = x-dt*v;
	v = 0.5 * (v + evalVel(y,evalVelData));
	y = x-dt*v;
	return y;
}

// trace forward in time (fully Lagrangian)
inline BfastVector3 bfastTraceForward(const BfastVector3 &x, BfastVector3 (*evalVel)(const BfastVector3 &x, void *data),
																			BfastReal dt, void *evalVelData=NULL) {
	BfastVector3 v = evalVel(x,evalVelData);
	BfastVector3 y = x+0.5*dt*v;
	v = evalVel(y,evalVelData);
	y = x+dt*v;
	return y;
}

#if 0
inline int bfastHash(int i, int j, int k) {
	return (i | j<<10 | k<<20);
}

inline int bfastHash(int *i) {
	return (i[0] | i[1]<<10 | i[2]<<20);
}

inline int bfastHash(BfastInt3 i) {
	return (i[0] | i[1]<<10 | i[2]<<20);
}
#endif 

inline BfastTri &BfastTri::operator=(const BfastTri &that) {
	this->indices[0] = that.indices[0];
	this->indices[1] = that.indices[1];
	this->indices[2] = that.indices[2];
	return (*this);
};

struct eqUnorderedInt3 {
  bool operator()(const BfastInt3 &i1, const BfastInt3 &i2) const {
    return ((i1[0] == i2[0] && i1[1] == i2[1] && i1[2] == i2[2]) ||
	    (i1[1] == i2[0] && i1[0] == i2[1] && i1[2] == i2[2]) ||
	    (i1[2] == i2[0] && i1[1] == i2[1] && i1[0] == i2[2]) ||
	    (i1[0] == i2[0] && i1[2] == i2[1] && i1[1] == i2[2]) ||
	    (i1[1] == i2[0] && i1[2] == i2[1] && i1[0] == i2[2]) ||
	    (i1[2] == i2[0] && i1[0] == i2[1] && i1[1] == i2[2]));
  }
};

struct eqOrderedInt3 {
	bool operator()(const BfastInt3 &i1, const BfastInt3 &i2) const {
		return (i1[0] == i2[0] && i1[1] == i2[1] && i1[2] == i2[2]);
	};
};

struct hashInt3 {
	size_t operator()(const BfastInt3 &i) const {
		return (i[0] ^ i[1] ^ i[2]);
	}
};
typedef HASH_MAP<BfastInt3,int,hashInt3,eqOrderedInt3> Int3ToIntMap;
typedef HASH_MAP<BfastInt3,double,hashInt3,eqOrderedInt3> Int3ToDoubleMap;

#endif

//-------------------------------------------------------------------
//-------------------------------------------------------------------
//
// RCS Revision History
//
// $Log: bfastUtil.H,v $
// Revision 1.11  2007/07/31 20:17:38  adamwb
// removed warnings
//
// Revision 1.10  2007/07/26 00:19:51  adamwb
// major updates to stl and faster hash_map usage
//
// Revision 1.9  2007/06/16 01:05:26  adamwb
// *** empty log message ***
//
// Revision 1.8  2007/01/22 14:04:16  adamwb
// collision object updates
//
// Revision 1.7  2006/12/23 03:09:30  adamwb
// ZZZZ
// 
//
// Revision 1.6  2006/06/16 21:07:14  adamb
// new version
//
// Revision 1.5  2006/01/15 01:43:52  adamb
// added rebuildTree
//
// Revision 1.4  2005/12/29 01:59:38  adamb
// Added void * to all passed in functions
//
// Revision 1.3  2005/11/08 19:45:05  adamb
// forced checkin
//
// Revision 1.2  2005/10/19 02:07:35  adamb
// seems to work
//
// Revision 1.1.1.1  2005/09/06 22:45:12  adamb
// imported sources
//
// Revision 1.1  2005/08/10 01:33:25  adamb
// forced
//
// Revision 1.3  2005/07/20 01:46:58  adamb
// forced checkin of slc code before changes to hash function
//
// Revision 1.2  2005/06/10 20:08:47  adamb
// Various bug fixes, added iuc and ilc to cell output, rebuild vertMap after readining in a surface, fixed argc checking in the drivers.
//
// Revision 1.1.1.1  2005/05/25 05:40:59  adamb
// Initial Revision
//
//
//-------------------------------------------------------------------
//-------------------------------------------------------------------
