// at initialization
SparseSymMatrix<double> A;

void mult (const SparseSymMatrix<double> &A, double *v, double *w) {
	A.mvmul(v,w);
}

int init() {
	std::vector<std::vector<int> > vertexToEdgeMap;
	std::vector<SimpleEdge> edgeToVertexMap;
	vertexToEdgeMap.resize(nodes.size());
	for (t=tets.begin(), i=0; t!=tets.end(); t++, i++) {
		for (int a=0; a<4; a++) {
			for (int b=a+1; b<4; b++) {
				bool found = false;
				for (unsigned int i=0; i<vertexToEdgeMap[(*t)[a]].size() && !found; i++) {
					if ((edgeToVertexMap[vertexToEdgeMap[(*t)[a]][i]][0] == (*t)[a] &&
							 edgeToVertexMap[vertexToEdgeMap[(*t)[a]][i]][1] == (*t)[b]) ||
							(edgeToVertexMap[vertexToEdgeMap[(*t)[a]][i]][0] == (*t)[b] &&
							 edgeToVertexMap[vertexToEdgeMap[(*t)[a]][i]][1] == (*t)[a]))
						found = true;
				}
				if (!found) {
					SimpleEdge e((*t)[a],(*t)[b]);
					int index = edgeToVertexMap.size();
					edgeToVertexMap.push_back(e);
					vertexToEdgeMap[(*t)[a]].push_back(index);
					vertexToEdgeMap[(*t)[b]].push_back(index);
				}
			}
		}
	}
	A.dim = 3*nodes.size();
	A.blocks = 0;
	for (unsigned int i=0; i<nodes.size(); i++) {
		A.blocks += vertexToEdgeMap[i].size();
	}
	A.blocks /= 2;
	A.blocks += nodes.size();
	A.rows = new int[A.blocks];
	A.cols = new int[A.blocks];
	A.values = new double[9*A.blocks];
	unsigned int b=0,c;
	for (unsigned int r=0; r<nodes.size(); r++) {
		A.rows[b] = r;
		A.cols[b] = r;
		b++;
		for (unsigned int e=0; e<vertexToEdgeMap[r].size(); e++) {
			if ((c = edgeToVertexMap[vertexToEdgeMap[r][e]][0]) == r) {
				c = edgeToVertexMap[vertexToEdgeMap[r][e]][1];
			}
			if (r > c) continue;
			A.rows[b] = r;
			A.cols[b] = c;
			b++;
		}
	}
}

// Code for every timestep
void timestep() {
	double *Ap = A.values;
	double dts = dt*dt;
	for (int i=0; i<9*A.blocks; i++,Ap++) *Ap = 0.0;

	for (std::vector<SimpleTet>::iterator t=tets.begin(); t!=tets.end(); t++) {
		// compute stress, etc.
		double K[12][12];
		double D[12][12];
		for (unsigned int i=0; i<12; i++) {
			for (unsigned int j=0; j<12; j++) {
				K[i][j] = 0.0;
				D[i][j] = 0.0;
			}
		}
		for (unsigned int a=0; a<3; a++) {
			for (unsigned int i=0; i<3; i++) {
				for (unsigned int e=0; e<3; e++) {
					for (unsigned int p=0; p<3; p++) {
						for (unsigned int j=0; j<3; j++) {
							for (unsigned int k=0; k<3; k++) {
								if (p == i) {
									double value = t->vol * t->beta(e,j)*t->beta(a,k)*t->stress(k,j);
									K[3*(a+1)+i][3*(e+1)+p] += value;
									K[i][p] += value;
									K[3*(a+1)+i][p] -= value;
									K[i][3*(e+1)+p] -= value;
								}
								for (unsigned int b=0; b<3; b++) {
									for (unsigned int d=0; d<3; d++) {
										double value = 
											lambda*t->X(i,b)*t->beta(b,k)*t->beta(a,k)*
											t->beta(e,j)*t->X(p,d)*t->beta(d,j) +
											
											mu*t->X(i,b)*t->beta(b,j)*t->beta(a,k)*
											t->beta(e,j)*t->X(p,d)*t->beta(d,k) +
											
											mu*t->X(i,b)*t->beta(b,j)*t->beta(a,k)*
											t->X(p,d)*t->beta(d,j)*t->beta(e,k);
										value *= t->vol;
										K[3*(a+1)+i][3*(e+1)+p] += value;
										K[i][p] += value;
										K[3*(a+1)+i][p] -= value;
										K[i][3*(e+1)+p] -= value;
										value =
											phi*t->X(i,b)*t->beta(b,k)*t->beta(a,k)*
											t->beta(e,j)*t->X(p,d)*t->beta(d,j) +
											
											psi*t->X(i,b)*t->beta(b,j)*t->beta(a,k)*
											t->beta(e,j)*t->X(p,d)*t->beta(d,k) +
											
											psi*t->X(i,b)*t->beta(b,j)*t->beta(a,k)*
											t->X(p,d)*t->beta(d,j)*t->beta(e,k);
										D[3*(a+1)+i][3*(e+1)+p] += value;
										D[i][p] += value;
										D[3*(a+1)+i][p] -= value;
										D[i][3*(e+1)+p] -= value;
									}
								}
							}
						}
					}
				}
			}
		}
		//for (unsigned int i=0; i<12; i++) {
		//for (unsigned int j=0; j<12; j++) {
		//	cout<<K[i][j]<<" ";
		//}
		//}
		for (unsigned int i=0; i<4; i++) {
			for (unsigned int j=i; j<4; j++) {
				int r = (*t)[i], c = (*t)[j];
				Ap = A.Find(3*r,3*c);
				if (Ap == NULL) cout<<"major problem"<<endl;
				for (unsigned int k=0; k<3; k++) {
					for (unsigned int l=0; l<3; l++) {
						*Ap += dt*D[3*i+k][3*j+l] + dts*K[3*i+k][3*j+l];
						if (r<=c) Ap++;
						else Ap+=3;
					}
					if (r > c) Ap-=8;
				}
			}
		}
		for (unsigned int i=0; i<4; i++) {
			for (unsigned int j=0; j<3; j++) {
				for (unsigned int k=0; k<4; k++) {
					for (unsigned int l=0; l<3; l++) {
						nodes[(*t)[i]].frc[j] += dt*K[3*i+j][3*k+l]*nodes[(*t)[k]].vel[l];
					}
				}
			}
		}
	}

	Ap = A.values;
	for (int i=0; i<A.blocks; i++, Ap+=9) {
		if (A.rows[i] == A.cols[i]) {
			*Ap += nodes[A.rows[i]].m;
			*(Ap+4) += nodes[A.rows[i]].m;
			*(Ap+8) += nodes[A.rows[i]].m;
		}
	}
	double eps = 1e-10;
	double *dv = new double[3*nodes.size()];
	double *b = new double[3*nodes.size()];
	
	for (unsigned int i=0; i<nodes.size(); i++) {
		b[3*i+0] = dt*nodes[i].frc[0];
		b[3*i+1] = dt*nodes[i].frc[1];
		b[3*i+2] = dt*nodes[i].frc[2];
		dv[3*i+0] = 0;
		dv[3*i+1] = 0;
		dv[3*i+2] = 0;
		//cout<<b[3*i+0]<<" "<<b[3*i+1]<<" "<<b[3*i+2]<<endl;
	}

	cghs(3*nodes.size(),A,b,dv,eps,false);

	for (unsigned int i=0; i<nodes.size(); i++) {
		//cout<<"dv["<<i<<"]:"<<dv[3*i+0]<<" "<<dv[3*i+1]<<" "<<dv[3*i+2]<<endl;
		nodes[i].vel[0] += dv[3*i+0];
		nodes[i].vel[1] += dv[3*i+1];
		nodes[i].vel[2] += dv[3*i+2];
		nodes[i].pos += dt * nodes[i].vel;
	}
}
