
#ifndef __SparseSymMatrix_hh__
#define __SparseSymMatrix_hh__

#include <iostream>
#include <fstream>

typedef unsigned int uint;

template <class FLOAT>
class SparseSymMatrix
{
public:

  SparseSymMatrix (int dim, int blocks, int* rows, int* cols, FLOAT* values);
  SparseSymMatrix();
  ~SparseSymMatrix ();
  
  void mvmul (FLOAT *a, FLOAT *b) const {
    for (int i=0; i<dim; i++) b[i] = 0.0;
    double *val = values;
    for (int block = 0; block < blocks; block++, val+=9) {
      b[3*rows[block]+0] += (*(val  ))*a[3*cols[block]+0]+
				(*(val+1))*a[3*cols[block]+1]+(*(val+2))*a[3*cols[block]+2];
      b[3*rows[block]+1] += (*(val+3))*a[3*cols[block]+0]+
				(*(val+4))*a[3*cols[block]+1]+(*(val+5))*a[3*cols[block]+2];
      b[3*rows[block]+2] += (*(val+6))*a[3*cols[block]+0]+
				(*(val+7))*a[3*cols[block]+1]+(*(val+8))*a[3*cols[block]+2];
      if (rows[block] != cols[block]) {
				b[3*cols[block]+0] += (*(val  ))*a[3*rows[block]+0]+
					(*(val+3))*a[3*rows[block]+1]+(*(val+6))*a[3*rows[block]+2];
				b[3*cols[block]+1] += (*(val+1))*a[3*rows[block]+0]+
					(*(val+4))*a[3*rows[block]+1]+(*(val+7))*a[3*rows[block]+2];
				b[3*cols[block]+2] += (*(val+2))*a[3*rows[block]+0]+
					(*(val+5))*a[3*rows[block]+1]+(*(val+8))*a[3*rows[block]+2];
      }
    }
  }

  inline FLOAT* Find(int row, int col);
  inline void getDiagonal(FLOAT *d);

  int dim;
  int blocks;
  int* rows;
  int* cols;
  double* values;
};

template <class FLOAT>
std::ostream &operator<<(std::ostream &out, const SparseSymMatrix<FLOAT> &A) {
  int i;
  out<<A.dim<<std::endl;
  out<<A.blocks<<std::endl;
  for (i=0; i<A.blocks; i++)
    out<<A.rows[i]<<" ";
  out<<std::endl;
  for (i=0; i<A.blocks; i++)
    out<<A.cols[i]<<" ";
  out<<std::endl;
  for (i=0; i<9*A.blocks; i++)
    out<<A.values[i]<<" ";
  out<<std::endl;
  return out;
}


template <class FLOAT>
std::istream &operator>>(std::istream &in, SparseSymMatrix<FLOAT> &A){
  int i,magic;

  if (A.rows) delete [] A.rows;
  if (A.cols) delete [] A.cols;
  if (A.values) delete [] A.values;

  in>>A.dim;
  in>>A.blocks;
  A.rows = new int[A.blocks];
  for (i=0; i<A.blocks; i++)
    in>>A.rows[i];
  A.cols = new int[A.blocks];
  for (i=0; i<A.blocks; i++)
    in>>A.cols[i];
  A.values = new double[9*A.blocks];
  for (i=0; i<9*A.blocks; i++)
    in>>A.values[i];
  return in;
}
  
template <class FLOAT>
SparseSymMatrix<FLOAT>::SparseSymMatrix ()
{
    this->dim = 0;
    this->blocks = 0;
    this->rows = NULL;
    this->cols = NULL;
    this->values = NULL;
}

template <class FLOAT>
SparseSymMatrix<FLOAT>::SparseSymMatrix (int dim, int blocks, int* rows, 
					 int* cols, FLOAT* values)
{
    this->dim = dim;
    this->blocks = blocks;
    this->rows = rows;
    this->cols = cols;
    this->values = values;
}

template <class FLOAT>
SparseSymMatrix<FLOAT>::~SparseSymMatrix ()
{
  
  if (rows) delete [] rows;
  if (cols) delete [] cols;
  if (values) delete [] values;
}


template<class FLOAT> 
inline FLOAT *SparseSymMatrix<FLOAT>::Find(int row,int col) {
  int i=blocks>>1,j=blocks>>1;
  int r=row/3, c=col/3;
  if (r < c) { 
    while (rows[i] != r) {
      if (j>1) j >>= 1;
      if (rows[i] < r) i+=j;
      else i-=j;
    }
    for (j=i; rows[j] == r; j++) {
      if (cols[j] == c) return &(values[9*j+3*(row%3)+col%3]);
    }
    for (j=i-1; rows[j] == r; j--) {
      if (cols[j] == c) return &(values[9*j+3*(row%3)+col%3]);
    }
  } else {
    while (rows[i] != c) {

      if (j>1) j >>= 1;
      if (rows[i] < c) i+=j;
      else i-=j;
    }
    for (j=i; rows[j] == c; j++) {
      if (cols[j] == r) return &(values[9*j+3*(col%3)+row%3]);
    }
    for (j=i-1; rows[j] == c; j--) {
      if (cols[j] == r) return &(values[9*j+3*(col%3)+row%3]);
    }
  }
  
  return NULL;

  for (i=0; i<blocks; i++) {
    if (rows[i] == row/3 && cols[i] == col/3) 
      return &(values[9*i+3*(row%3)+col%3]);
    if (rows[i] == col/3 && cols[i] == row/3)
      return &(values[9*i+3*(col%3)+row%3]);
  }

  for (i=0; i<blocks; i++) {
    if (rows[i] == row/3 && cols[i] == col/3) 
      return &(values[9*i+3*(row%3)+col%3]);
    if (rows[i] == col/3 && cols[i] == row/3)
      return &(values[9*i+3*(col%3)+row%3]);
  }

  return NULL;
}

inline double *Find(int row,int col,int blocks,int *rows,int *cols,double *values) {
  int i;
  for (i=0; i<blocks; i++) {
    if (rows[i] == row/3 && cols[i] == col/3) 
      return &(values[9*i+3*(row%3)+col%3]);
    if (rows[i] == col/3 && cols[i] == row/3)
      return &(values[9*i+3*(col%3)+row%3]);
  }

  for (i=0; i<blocks; i++) {
    if (rows[i] == row/3 && cols[i] == col/3) 
      return &(values[9*i+3*(row%3)+col%3]);
    if (rows[i] == col/3 && cols[i] == row/3)
      return &(values[9*i+3*(col%3)+row%3]);
  }

  return NULL;
}

template<class FLOAT> 
inline void SparseSymMatrix<FLOAT>::getDiagonal(FLOAT *d) {
  FLOAT* dp = d;
  FLOAT* vp = values;
  for (int i=0; i<blocks; i++, vp+=9) {
    if (rows[i] == cols[i]) {
      *(dp++) = *vp;
      *(dp++) = *(vp+4);
      *(dp++) = *(vp+8);
    }
  }
}


#endif // __SparseSymMatrix_hh__

